#!/usr/bin/env python
import sys
from os.path import abspath, dirname

import django
from django.conf import settings


sys.path.insert(0, abspath(dirname(__file__)))


if not settings.configured:
    settings.configure(
        INSTALLED_APPS=(
            'django.contrib.contenttypes',
            'django.contrib.sessions',
            'django.contrib.messages',
            'django.contrib.auth',
            'relatives',
            'relatives.tests',
            'django.contrib.admin',
        ),
        DATABASES={
            'default': {
                'NAME': ':memory:',
                'ENGINE': 'django.db.backends.sqlite3',
            }
        },
        MIDDLEWARE=[
            'django.contrib.sessions.middleware.SessionMiddleware',
            'django.middleware.common.CommonMiddleware',
            'django.contrib.auth.middleware.AuthenticationMiddleware',
            'django.contrib.messages.middleware.MessageMiddleware',
        ],
        ROOT_URLCONF='relatives.tests.urls',
        SECRET_KEY='secret',
        STATIC_URL='/static/',
        DEFAULT_AUTO_FIELD='django.db.models.BigAutoField',
        TEMPLATES=[
            {
                'BACKEND': 'django.template.backends.django.DjangoTemplates',
                'APP_DIRS': True,
                'OPTIONS': {
                    "context_processors": [
                        'django.contrib.auth.context_processors.auth',
                        'django.contrib.messages.context_processors.messages',
                        'django.template.context_processors.request',
                    ]
                }
            },
        ],
    )


def runtests():
    if hasattr(django, 'setup'):
        django.setup()
    try:
        from django.test.runner import DiscoverRunner
        runner_class = DiscoverRunner
        test_args = ['relatives.tests']
    except ImportError:
        from django.test.simple import DjangoTestSuiteRunner
        runner_class = DjangoTestSuiteRunner
        test_args = ['tests']
    failures = runner_class(failfast=False).run_tests(test_args)
    sys.exit(failures)


def run_management_command(arguments):
    if hasattr(django, 'setup'):
        django.setup()
    from django.core.management import execute_from_command_line
    execute_from_command_line(arguments)


if __name__ == "__main__":
    if sys.argv[1:]:
        run_management_command(sys.argv)
    else:
        runtests()



from __future__ import unicode_literals
from django.test import TestCase
from django.template.loader import render_to_string
from django.urls import reverse
from django.contrib.auth.models import User
from django.contrib.contenttypes.models import ContentType
from django.core.cache import cache

from relatives.utils import object_link, object_edit_link
from .models import (Pirate, Pet, Ship, Sailor, Movie, Actor, NotInAdmin,
                     Shape, Something, Book, Image, Journal)


class ObjectEditLinkTest(TestCase):

    def test_default(self):
        ship = Ship.objects.create(id=1, name="Star of India")
        self.assertEqual(
            object_edit_link()(ship),
            '<a href="/adm/tests/ship/1/change/">Star of India</a>',
        )
        pirate = Pirate.objects.create(id=1, name="Lowell Taylor")
        self.assertEqual(object_edit_link()(pirate), "Lowell Taylor")

    def test_custom_edit_text(self):
        ship = Ship.objects.create(id=1, name="Star of India")
        self.assertEqual(object_edit_link("Go There")(ship),
                         '<a href="/adm/tests/ship/1/change/">Go There</a>')

    def test_default_blank_text(self):
        pirate = Pirate.objects.create(id=1, name="Lowell Taylor")
        self.assertEqual(object_edit_link("Edit")(pirate), "")

    def test_custom_edit_and_blank_text(self):
        ship = Ship.objects.create(id=1, name="Star of India")
        pirate = Pirate.objects.create(id=1, name="Lowell Taylor")
        self.assertEqual(object_edit_link("Go There", "N/A")(ship),
                         '<a href="/adm/tests/ship/1/change/">Go There</a>')
        self.assertEqual(object_edit_link("Go There", "N/A")(pirate), "N/A")


class ObjectLinkTest(TestCase):

    def test_no_admin_url(self):
        pirate = Pirate.objects.create(id=1, name="Lowell Taylor")
        self.assertEqual(object_link(pirate), "Lowell Taylor")

    def test_with_primary_key(self):
        ship = Ship.objects.create(id=1, name="Star of India")
        self.assertEqual(
            object_link(ship),
            '<a href="/adm/tests/ship/1/change/">Star of India</a>',
        )

    def test_no_primary_key(self):
        ship = Ship(name="Star of India")
        self.assertEqual(object_link(ship), "Star of India")


class TemplateFilterTest(TestCase):
    def setUp(self):
        self.user = User.objects.create_superuser('u', 'u@example.com', 'pass')

    def login(self):
        self.client.login(username=self.user.username, password='pass')

    def test_foreign_key(self):
        self.login()
        ship = Ship.objects.create(id=1, name="Star of India")
        sailor = Sailor.objects.create(name="John Ford", ship=ship)
        response = self.client.get(reverse('admin:tests_sailor_change',
                                           args=[sailor.id]))
        self.assertIn(b'<a href="/adm/tests/ship/1/change/">Star of India</a>',
                      response.content)

    def test_property_and_modeladmin_method(self):
        self.login()
        circle = Shape.objects.create(id=1, name="Circle")
        response = self.client.get(reverse('admin:tests_shape_change',
                                           args=[circle.id]))
        self.assertIn(b'<p>Circle</p>', response.content)
        self.assertIn(b'<p>circle</p>', response.content)
        self.assertIn(b'<p>CIRCLE</p>', response.content)

    def test_no_foreign_key(self):
        self.login()
        ship = Ship.objects.create(id=1, name="Star of India")
        response = self.client.get(reverse('admin:tests_ship_change',
                                           args=[ship.id]))
        self.assertIn(b'<p>Star of India</p>', response.content)

    def test_foreign_key_without_admin_url(self):
        self.login()
        pirate = Pirate.objects.create(id=1, name="Kristi Bell")
        pet = Pet.objects.create(owner=pirate)
        response = self.client.get(reverse('admin:tests_pet_change',
                                           args=[pet.id]))
        self.assertIn(b'Kristi Bell', response.content)
        self.assertNotIn(b'Kristi Bell</a>', response.content)
        self.assertNotIn(b'Kristi Bell</option>', response.content)

    def test_nullable_foreign_key(self):
        self.login()
        sailor = Sailor.objects.create(name="John Ford")
        response = self.client.get(reverse('admin:tests_sailor_change',
                                           args=[sailor.id]))
        self.assertIn(b'<p>-</p>', response.content)

    def test_deleted_foreign_key(self):
        self.login()
        ship = Ship.objects.create(id=1, name="Star of India")
        sailor = Sailor.objects.create(name="John Ford", ship=ship)
        ship.delete()  # Sailor won't get deleted
        response = self.client.get(reverse('admin:tests_sailor_change',
                                           args=[sailor.id]))
        self.assertIn(b'<p>-</p>', response.content)

    def test_add_form_for_non_nullable_fk(self):
        self.login()
        response = self.client.get(reverse('admin:tests_pet_add'))
        self.assertIn(b'<p>-</p>', response.content)

    def test_generic_foreign_key_present(self):
        book = Book.objects.create(name="Django")
        image = Image.objects.create(content_object=book)
        self.login()
        response = self.client.get(
            reverse('admin:tests_image_change', args=[image.id])
        )
        self.assertIn(f'/adm/tests/book/{book.id}', response.rendered_content)
        self.assertIn(f'Book object ({book.id})', response.rendered_content)
        self.assertNotIn('<p>None</p>', response.rendered_content)

    def test_generic_foreign_key_not_present(self):
        image = Image.objects.create()
        self.login()
        response = self.client.get(
            reverse('admin:tests_image_change', args=[image.id])
        )
        self.assertIn('<p>None</p>', response.rendered_content)
        self.assertNotIn('Book object', response.rendered_content)


class RelatedObjectsTagTest(TestCase):
    def test_foreign_keys(self):
        ship = Ship.objects.create(id=1, name="Star of India")
        body = render_to_string('related_objects_fk_test.html', {'obj': ship})
        self.assertEqual(body.strip(),
                         '<a href="/adm/tests/sailor/?ship=1">Sailors</a>')

    def test_no_admin_url(self):
        thing = Something.objects.create()
        NotInAdmin.objects.create(id=1, fk=thing)
        body = render_to_string('related_objects_fk_test.html', {'obj': thing})
        self.assertEqual(body.strip(), '')

    def test_many_to_many(self):
        movie = Movie.objects.create(id=1, name="Yojimbo")
        actor = Actor.objects.create(name="Tatsuya Nakadai")
        actor.movies.add(movie)
        body = render_to_string('related_objects_m2m_test.html',
                                {'obj': movie})
        self.assertEqual(body.strip(),
                         '<a href="/adm/tests/actor/?movies=1">Actors</a>')

    def test_reverse_many_to_many(self):
        movie = Movie.objects.create(id=1, name="Yojimbo")
        actor = Actor.objects.create(name="Tatsuya Nakadai")
        actor.movies.add(movie)
        body = render_to_string('related_objects_m2m_test.html',
                                {'obj': actor})
        self.assertEqual(body.strip(), '')

    def test_generic_foreign_key_not(self):
        book = Book.objects.create(name="Django")
        body = render_to_string('related_objects_generic_test.html',
                                {'obj': book})
        self.assertEqual(body.strip(), '')

    def test_generic_foreign_key_present(self):
        book = Book.objects.create(name="Django")
        Image.objects.create(content_object=book)
        ct_pk = ContentType.objects.get_for_model(book).pk
        exp = '<a href="/adm/tests/image/?ct={0}&amp;obj_id={1}">Images</a>'\
            .format(ct_pk, book.pk)
        body = render_to_string('related_objects_generic_test.html',
                                {'obj': book})
        self.assertEqual(body.strip(), exp)

    def test_generic_relation_and_gfk_present(self):
        journal = Journal.objects.create(name="Django")
        Image.objects.create(content_object=journal)
        ct_pk = ContentType.objects.get_for_model(journal).pk
        exp = '<a href="/adm/tests/image/?ct={0}&amp;obj_id={1}">Images</a>'\
            .format(ct_pk, journal.pk)
        body = render_to_string('related_objects_generic_test.html',
                                {'obj': journal})
        self.assertEqual(body.strip(), exp)

    def test_with_removed_model(self):
        cache.clear()
        journal = Journal()
        ContentType.objects.create(model='gone', app_label='gone')
        body = render_to_string('related_objects_generic_test.html',
                                {'obj': journal})
        self.assertEqual(body.strip(), '')



from django.utils.html import conditional_escape
from django.utils.safestring import mark_safe


def format_html(format_string, *args, **kwargs):
    """
    Similar to str.format, but passes all arguments through conditional_escape,
    and calls 'mark_safe' on the result. This function should be used instead
    of str.format or % interpolation to build up small HTML fragments.
    """
    args_safe = (conditional_escape(a) for a in args)
    kwargs_safe = {
        k: conditional_escape(v)
        for (k, v) in kwargs.items()
    }
    return mark_safe(format_string.format(*args_safe, **kwargs_safe))



from __future__ import unicode_literals
from django.utils.encoding import smart_text
from .compat import format_html
from django.contrib.contenttypes.models import ContentType
from django.urls import reverse, NoReverseMatch
from django.core.cache import cache
from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey


def get_admin_url(obj):
    """Return admin URL for given object (raise NoReverseMatch on error)"""
    options = obj._meta.app_label, obj._meta.model_name
    return reverse('admin:%s_%s_change' % options, args=[obj.pk])


def object_edit_link(edit_text=None, blank_text=None):

    """
    Return function that takes an object and returns admin link to object

    Arguments:

    - ``edit_text`` is displayed in link text
    - ``blank_text`` is displayed in unlinked text (when no admin link)

    ``edit_text`` defaults to the object's unicode representation and
    ``blank_text`` defaults to the object's unicode representation if
    ``edit_text`` is None and an empty string otherwise
    """

    def object_link(obj):
        "Return admin link to given object or blank text if no link"
        link_text = smart_text(obj) if edit_text is None else edit_text
        try:
            if obj.pk:
                return format_html('<a href="{0}">{1}</a>', get_admin_url(obj),
                                   link_text)
        except NoReverseMatch:
            pass
        if blank_text is None:
            if edit_text is None:
                return link_text
            else:
                return ""
        else:
            return blank_text
    object_link.__name__ = str("")
    object_link.allow_tags = True
    return object_link


object_link = object_edit_link()
object_link.allow_tags = True
object_link.__doc__ += "\n\nEquivalent to object_edit_link()(obj)"


class RelatedObject(object):
    '''Generates fake django RelatedObject
    '''

    def __init__(self, field, ct_pk):
        self.field = field
        self.field.name = self.generate_field_name(field, ct_pk)
        self.model = self.field.model
        self.name = self.generate_name(field)

    @staticmethod
    def generate_name(field):
        return ':'.join([field.model._meta.app_label,
                        field.model._meta.model_name])

    @staticmethod
    def generate_field_name(field, ct_pk):
        return ''.join(
            map(str, [field.ct_field, '=', ct_pk, '&', field.fk_field]))


class GenericObjects(object):
    '''
    Search GenericForeignKey over all models
    and returns related objects if has relations with given object.
    '''

    def __init__(self, _object):
        self.obj = _object
        self.ct_pk = ContentType.objects.get_for_model(_object).pk
        self.cache_key = getattr(settings, 'RELATIVES_CACHE_KEY',
                                 'relatives_cache')
        self.cache_time = getattr(settings, 'RELATIVES_CACHE_TIME',
                                  60 * 60 * 24)
        self.generic_objects = []

    def get_generic_objects(self):
        try:
            self._generic_fields_cache
        except AttributeError:
            self._fill_generic_fields_cache()
        for generic_field in self._generic_fields_cache:
            params = {generic_field.ct_field: self.ct_pk,
                      generic_field.fk_field: self.obj.pk}
            if generic_field.model.objects.filter(**params).exists():
                generic_object = RelatedObject(generic_field, self.ct_pk)
                self.generic_objects.append(generic_object)
        return self.generic_objects

    def _fill_generic_fields_cache(self):
        self._generic_fields_cache = cache.get(self.cache_key)
        if self._generic_fields_cache is not None:
            return
        self._generic_fields_cache = []
        for content_type in ContentType.objects.all():
            model_class = content_type.model_class()
            if not model_class:
                continue  # Model for content type doesn't exist anymore
            self._generic_fields_cache += [
                field
                for field in model_class._meta.private_fields
                if isinstance(field, GenericForeignKey)
            ]
        cache.set(self.cache_key, self._generic_fields_cache, self.cache_time)



from django.contrib import admin

from . import models


class ShapeAdmin(admin.ModelAdmin):
    readonly_fields = ['name', 'lower_name', 'upper_name']

    def upper_name(self, obj):
        return obj.name.upper()


admin.site.register(models.Shape, ShapeAdmin)


class ShipAdmin(admin.ModelAdmin):
    change_form_template = 'relatives/change_form.html'
    readonly_fields = ['name']


admin.site.register(models.Ship, ShipAdmin)


class SailorAdmin(admin.ModelAdmin):
    readonly_fields = ['ship']


admin.site.register(models.Sailor, SailorAdmin)


class PetAdmin(admin.ModelAdmin):
    readonly_fields = ['owner']


admin.site.register(models.Pet, PetAdmin)


class ImageAdmin(admin.ModelAdmin):
    readonly_fields = ['content_object']


admin.site.register(models.Image, ImageAdmin)


admin.site.register(models.Movie)
admin.site.register(models.Actor)
admin.site.register(models.Book)
admin.site.register(models.Journal)



from pathlib import Path
from setuptools import setup, find_packages
import relatives


setup(
    name="django-relatives",
    version=relatives.__version__,
    author="Trey Hunner",
    author_email="trey@treyhunner.com",
    url="https://github.com/treyhunner/django-relatives",
    description="Utilities for linking to related objects in Django admin",
    long_description='\n\n'.join((
        Path('README.rst').read_text(),
        Path('CHANGES.rst').read_text(),
    )),
    packages=find_packages(),
    include_package_data=True,
    install_requires=['Django >= 1.11'],
    tests_require=['Django >= 1.11'],
    test_suite='runtests.runtests',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Framework :: Django',
    ],
)



from django.template import Library
from django.urls import reverse, NoReverseMatch
from django.utils.encoding import smart_text
from django.utils.safestring import mark_safe
from django.contrib.admin.utils import lookup_field
from django.core.exceptions import ObjectDoesNotExist

from ..utils import get_admin_url, GenericObjects


register = Library()


@register.filter
def contents_or_fk_link(field):
    """
    Return field contents or link to related object if foreign key field

    Example Usage::

        {% load relatives %}
        {{ field|contents_or_fk_link }}
    """
    contents = field.contents()
    field_name = field.field['field']
    obj = field.form.instance
    if not isinstance(field_name, str):
        return contents
    try:
        related_obj = getattr(obj, field_name)
    except (AttributeError, ObjectDoesNotExist):
        return contents
    else:
        model_field, _, _ = lookup_field(field_name, obj, field.model_admin)
        if ((model_field is None or model_field.remote_field)
                and hasattr(related_obj, '_meta')):
            try:
                return mark_safe('<a href="%s">%s</a>' %
                                 (get_admin_url(related_obj), contents))
            except NoReverseMatch:
                pass
        return contents


@register.simple_tag
def related_objects(obj):
    """
    Return list of objects related to the given model instance

    Example Usage::

        {% load relatives %}
        {% related_objects obj as related_objects %}
        {% for related_obj in related_objects %}
            <a href="{{ related_obj.url }}">{{ related_obj.plural_name }}</a>
        {% endfor %}
    """
    object_list = []
    all_related_objects = [
        field
        for field in obj._meta.get_fields()
        if (field.one_to_many or field.one_to_one)
        and field.auto_created and not field.concrete
    ]
    all_related_m2m_objects = [
        field
        for field in obj._meta.get_fields(include_hidden=True)
        if field.many_to_many and field.auto_created
    ]
    related_objects = (all_related_objects
                       + all_related_m2m_objects
                       + GenericObjects(obj).get_generic_objects())
    for related in related_objects:
        try:
            to_model = getattr(related, 'related_model', related.model)
            url = reverse('admin:{0}_{1}_changelist'.format(
                to_model._meta.app_label,
                to_model._meta.model_name
            ))
        except NoReverseMatch:
            continue
        object_list.append({
            'plural_name': to_model._meta.verbose_name_plural,
            'url': smart_text('%s?%s=%s' % (url, related.field.name, obj.pk)),
        })
    return object_list



# flake8: noqa
# Generated by Django 3.0.2 on 2020-01-28 12:38

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Book',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name='Journal',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name='Movie',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=80)),
            ],
        ),
        migrations.CreateModel(
            name='Pirate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=80)),
            ],
        ),
        migrations.CreateModel(
            name='Ship',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=40)),
            ],
        ),
        migrations.CreateModel(
            name='Something',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', models.CharField(max_length=10)),
            ],
        ),
        migrations.CreateModel(
            name='Sailor',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=80)),
                ('ship', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='tests.Ship')),
            ],
        ),
        migrations.CreateModel(
            name='Pet',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('owner', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='tests.Pirate')),
            ],
        ),
        migrations.CreateModel(
            name='NotInAdmin',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('fk', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='tests.Something')),
            ],
        ),
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('obj_id', models.PositiveIntegerField(null=True)),
                ('ct', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='Actor',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=80)),
                ('movies', models.ManyToManyField(related_name='actors', to='tests.Movie')),
            ],
        ),
    ]
